from pandas import Series, concat


def calculate_peak_prominence(series: Series) -> Series:
    """
    Get peak prominence for each value in a numerical series
    """
    # Calculate the absolute value of each element (we're comparing absolute magnitudes)
    abs_series = series.abs()

    # Shift the series to get the previous and next values
    left_rolling = abs_series.shift(1)  # previous element
    right_rolling = abs_series.shift(-1)  # next element

    # For boundary cases, assume the value itself to avoid NaNs
    left_rolling.fillna(abs_series, inplace=True)
    right_rolling.fillna(abs_series, inplace=True)

    # Identify peaks by checking if the current value is larger than both neighbors
    is_peak = ((abs_series > left_rolling) & (abs_series > right_rolling)) | ((abs_series < left_rolling) & (abs_series < right_rolling))

    # Calculate prominence for peaks only
    prominence = Series(0, index=series.index)  # Initialize with zeros
    prominence[is_peak] = (
        abs_series[is_peak]
        - concat([left_rolling, right_rolling], axis=1).min(axis=1)[is_peak]
    )

    return prominence.abs()



"""
Minimum Validator

The code is licensed under the MIT license.
"""

from pandas import Series
from ..validator import Validator


def minimum(value: int | float) -> Series:
    """
    Numeric minimum
    """

    def _func(series: Series) -> Series:
        return series >= value

    return Validator(_func)



"""
Validator Class

The code is licensed under the MIT license.
"""

from inspect import isfunction, signature
from typing import Callable, Optional
from pandas import DataFrame, Series


class Validator:
    """
    Schema Column Validator
    """

    func: Optional[Callable] = None
    ignore_na = False

    def __init__(self, func: Callable, ignore_na=False):
        self.func = func
        self.ignore_na = ignore_na

    def run(self, series: Series, df: DataFrame, column: str) -> bool | Series:
        """
        Run validator

        Returns a bool series:
        True -> Check passed
        False -> Check failed
        """
        arg_count = len((signature(self.func)).parameters)
        args = [series, df, column]
        return self.func(*args[0:arg_count])


def apply_validator(validator: Validator, df: DataFrame, column: str) -> Series:
    """
    Validate a DataFrame's column using a validator
    """
    validator = validator() if isfunction(validator) else validator
    if validator.ignore_na:
        result = Series(data=True, index=df.index, dtype=bool)
        result.update(
            validator.run(
                df.loc[df[column].notnull()][column],
                df.loc[df[column].notnull()],
                column,
            )
        )
        return result.astype(bool)
    return validator.run(df[column], df, column)



"""
Regular Expression Validator

The code is licensed under the MIT license.
"""

from pandas import Series
from ..validator import Validator


def regex(pattern: str) -> Series:
    """
    Check if column values match regex pattern
    """

    def _func(series: Series) -> Series:
        return series.str.contains(pattern)

    return Validator(_func, ignore_na=True)



"""
Schema Class

The code is licensed under the MIT license.
"""

from copy import copy
from typing import List, Union
from pandas import DataFrame, to_numeric

from pulire.column import Column
from pulire.validator import apply_validator
from pulire.formatter import apply_formatter


class Schema:
    """
    DataFrame Schema
    """

    columns: List[Column]

    def __init__(self, columns: List[Column]):
        self.columns = columns

    @property
    def names(self) -> List[str]:
        """
        List of column names
        """
        return [col.name for col in self.columns]

    @property
    def dtypes(self) -> dict:
        """
        Dictionary of data types
        """
        return {col.name: col.dtype for col in self.columns}

    def fit(self, df: DataFrame) -> DataFrame:
        """
        Set data types, apply formatters and remove invalid data from a DataFrame
        """
        temp = copy(df)

        temp = self.purge(temp)
        temp = self.fill(temp)
        temp = self.format(temp)
        temp = self.clean(temp)

        return temp
    
    def purge(self, df: DataFrame) -> DataFrame:
        """
        Remove DataFrame columns which are not in the schema
        """
        columns = [col for col in self.names if col in df.columns]
        return df[columns]
    
    def fill(self, df: DataFrame, value = None) -> DataFrame:
        """
        Add missing schema columns to DataFrame
        """
        temp = copy(df)

        for col in self.names:
            if col not in df:
                temp[col] = value

        return temp

    def format(self, df: DataFrame) -> DataFrame:
        """
        Set data types and apply formatters to a DataFrame
        """
        temp = copy(df)

        # Set data types
        for col, dtype in self.dtypes.items():
            if col in temp:
                if dtype == "Int64":
                    temp[col] = to_numeric(temp[col]).round(0)
                temp[col] = temp[col].astype(dtype, errors="ignore")

        # Apply formatters
        for col in self.columns:
            if col.name in temp.columns:
                for formatter in col.formatters:
                    temp[col.name] = apply_formatter(formatter, temp, col.name)

        return temp

    def clean(self, df: DataFrame, fill=None) -> DataFrame:
        """
        Remove invalid data from a DataFrame
        """
        temp = copy(df)

        for col in self.columns:
            if col.name in temp.columns:
                for validator in col.validators:
                    test = apply_validator(validator, temp, col.name)
                    temp.loc[~test, col.name] = fill

        return temp

    def validate(self, df: DataFrame, lazy=False) -> None:
        """
        Raise error for failing validations
        """
        errors: List[str] = []

        for col in self.columns:
            if col.name in df.columns:
                for validator in col.validators:
                    test = apply_validator(validator, df, col.name)
                    if not test.all():
                        for index in test[~test].index:
                            value = df.loc[index, col.name]
                            error = f'Value "{value}" at index "{index}" in column "{col.name}" did not pass validation'
                            if lazy:
                                errors.append(error)
                            else:
                                raise ValueError(error)

        raise ValueError(*errors)

    def match(self, df: DataFrame) -> bool:
        """
        Check if a DataFrame is valid
        """
        for col in self.columns:
            if col.name in df.columns:
                for validator in col.validators:
                    test = apply_validator(validator, df, col.name)
                    if not test.all():
                        return False
        return True

    def __getitem__(self, key: str | List[str]) -> Union["Schema", Column]:
        """
        Select a subset of columns or a specific Column object by name
        """
        if isinstance(key, str):
            # Return the specific Column object for the given column name
            for col in self.columns:
                if col.name == key:
                    return col
            raise KeyError(f"Column '{key}' not found in schema.")
        elif isinstance(key, list):
            # Return a new Schema with a subset of columns
            cols = [col for col in self.columns if col.name in key]
            return Schema(cols)
        raise TypeError("Key must be a string or a list of strings.")


"""
Is-In Validator

The code is licensed under the MIT license.
"""

from pandas import Series
from ..validator import Validator


def isin(values: list) -> Series:
    """
    Require column value to be in a list of values
    """

    def _func(series: Series) -> Series:
        return series.isin(values)

    return Validator(_func, ignore_na=True)



"""
Maximum Difference Validator

The code is licensed under the MIT license.
"""

from pandas import Series
from ..validator import Validator


def max_diff(value: int | float) -> Series:
    """
    Maximum difference compared to previous value
    """

    def _func(series: Series) -> Series:
        result = Series(data=0, index=series.index)
        result.update(series.iloc[1:].diff().notnull().abs())
        return result <= value

    return Validator(_func)



"""
Pulire - A lightweight DataFrame validation library.

The code is licensed under the MIT license.
"""

__appname__ = "pulire"
__version__ = "0.0.8"

from . import validators, formatters
from .validator import Validator
from .formatter import Formatter
from .schema import Schema
from .column import Column

__all__ = ["validators", "formatters", "Validator", "Formatter", "Schema", "Column"]



"""
Required Validator

The code is licensed under the MIT license.
"""

from pandas import Series, notna
from ..validator import Validator


def required() -> Series:
    """
    Require the current column not to be null
    """
    return Validator(notna)



"""
Requires Validator

The code is licensed under the MIT license.
"""

from pandas import Series
from ..validator import Validator


def requires(column: str) -> Series:
    """
    Require another column not to be null
    """

    def _func(series: Series) -> Series:
        return series[column].notna()

    return Validator(_func)



"""
Formatter Class

The code is licensed under the MIT license.
"""

from inspect import isfunction, signature
from typing import Callable, Optional
from pandas import DataFrame, Series


class Formatter:
    """
    Schema Column Formatter
    """

    func: Optional[Callable] = None

    def __init__(self, func: Callable):
        self.func = func

    def run(self, series: Series, df: DataFrame, column: str) -> Series:
        """
        Format all values in a series
        """
        arg_count = len((signature(self.func)).parameters)
        args = [series, df, column]
        return self.func(*args[0:arg_count])


def apply_formatter(formatter: Formatter, df: DataFrame, column: str) -> Series:
    """
    Format a DataFrame's column using a formatter
    """
    formatter = formatter() if isfunction(formatter) else formatter
    return formatter.run(df[column], df, column)



"""
Decimals Formatter

The code is licensed under the MIT license.
"""

from pandas import Series
from ..formatter import Formatter


def decimals(digits: int) -> Series:
    """
    Round a series to the specified number of fractional digits
    """

    def _func(series: Series) -> Series:
        return series.round(digits)

    return Formatter(_func)



"""
Maximum Peak Prominence Validator

The code is licensed under the MIT license.
"""

from pandas import Series
from pulire.utils import calculate_peak_prominence
from ..validator import Validator


def max_peak(value: int | float) -> Series:
    """
    Maximum peak prominence of a values in a time series
    """

    def _func(series: Series) -> Series:
        prominence = calculate_peak_prominence(series)
        return prominence < value

    return Validator(_func)



"""
Greater Validator

The code is licensed under the MIT license.
"""

from pandas import Series, DataFrame
from ..validator import Validator


def greater(column: str) -> Series:
    """
    Require column to be greather than another one
    """

    def _func(series: Series, df: DataFrame, name: str) -> Series:
        result = Series(data=True, index=series.index)
        df = df[df[name].notnull() & df[column].notnull()]
        result.update(df[name] > df[column])
        return result.astype(bool)

    return Validator(_func)



"""
Schema Validators

The code is licensed under the MIT license.
"""

from .required import required
from .minimum import minimum
from .maximum import maximum
from .requires import requires
from .greater import greater
from .less import less
from .max_fall import max_fall
from .max_rise import max_rise
from .max_diff import max_diff
from .max_peak import max_peak
from .isin import isin
from .regex import regex

__all__ = [
    "required",
    "minimum",
    "maximum",
    "requires",
    "greater",
    "less",
    "max_fall",
    "max_rise",
    "max_diff",
    "max_peak",
    "isin",
    "regex",
]



"""
Column Class

The code is licensed under the MIT license.
"""

from typing import List

from pulire.formatter import Formatter
from pulire.validator import Validator


class Column:
    """
    Schema Column
    """

    name: str
    dtype: str
    annotations: List[Validator | Formatter]

    def __init__(
        self, name: str, dtype: str, annotations: List[Validator | Formatter] = []
    ):
        self.name = name
        self.dtype = dtype
        self.annotations = annotations

    @property
    def validators(self) -> List[Validator]:
        """
        List of validators
        """
        return [
            annotation
            for annotation in self.annotations
            if isinstance(annotation, Validator)
        ]

    @property
    def formatters(self) -> List[Formatter]:
        """
        List of formatters
        """
        return [
            annotation
            for annotation in self.annotations
            if isinstance(annotation, Formatter)
        ]
